@@ -513,15 +513,6 @@ panel_applet_create_menu (AppletInfo *info)
 				  G_CALLBACK (applet_remove_callback), info);
 		gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
 		gtk_widget_set_sensitive (menuitem, removable);
-
-		menuitem = gtk_check_menu_item_new ();
-		setup_menuitem (menuitem, GTK_ICON_SIZE_MENU, NULL, _("_Lock To Panel"), FALSE);
-		gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (menuitem),
-						locked);
-		g_signal_connect (menuitem, "toggled",
-				  G_CALLBACK (panel_applet_lock), info);
-		gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
-		gtk_widget_set_sensitive (menuitem, lockable);
 		
 		menuitem = gtk_image_menu_item_new ();
 		setup_menuitem (menuitem, GTK_ICON_SIZE_MENU, NULL, _("_Move"), FALSE);
@@ -536,6 +527,19 @@ panel_applet_create_menu (AppletInfo *info)
 		g_object_add_weak_pointer (G_OBJECT (menuitem),
 					   (gpointer *) &info->move_item);
 
+		menuitem = gtk_separator_menu_item_new ();
+		gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
+		gtk_widget_show (menuitem);
+
+		menuitem = gtk_check_menu_item_new ();
+		setup_menuitem (menuitem, GTK_ICON_SIZE_MENU, NULL, _("_Lock To Panel"), FALSE);
+		gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (menuitem),
+						locked);
+		g_signal_connect (menuitem, "toggled",
+				  G_CALLBACK (panel_applet_lock), info);
+		gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
+		gtk_widget_set_sensitive (menuitem, lockable);
+
 		added_anything = TRUE;
 	}
 
@@ -581,6 +585,8 @@ panel_applet_position_menu (GtkMenu   *menu,
 	GdkScreen      *screen;
 	int             menu_x = 0;
 	int             menu_y = 0;
+	int             pointer_x;
+	int             pointer_y;
 
 	g_return_if_fail (PANEL_IS_WIDGET (applet->parent));
 
@@ -589,6 +595,7 @@ panel_applet_position_menu (GtkMenu   *menu,
 	gtk_widget_size_request (GTK_WIDGET (menu), &requisition);
 
 	gdk_window_get_origin (applet->window, &menu_x, &menu_y);
+	gtk_widget_get_pointer (applet, &pointer_x, &pointer_y);
 
 	if (GTK_WIDGET_NO_WINDOW (applet)) {
 		menu_x += applet->allocation.x;
@@ -596,11 +603,19 @@ panel_applet_position_menu (GtkMenu   *menu,
 	}
 
 	if (PANEL_WIDGET (applet->parent)->orient == GTK_ORIENTATION_HORIZONTAL) {
+		if (requisition.width < pointer_x)
+			menu_x += MIN (pointer_x, applet->allocation.width - requisition.width);
+		menu_x = MIN (menu_x, gdk_screen_get_width (screen) - requisition.width);
+
 		if (menu_y > gdk_screen_get_height (screen) / 2)
 			menu_y -= requisition.height;
 		else
 			menu_y += applet->allocation.height;
 	} else {
+		if (requisition.height < pointer_y)
+			menu_y += MIN (pointer_y, applet->allocation.height - requisition.height);
+		menu_y = MIN (menu_y, gdk_screen_get_height (screen) - requisition.height);
+
 		if (menu_x > gdk_screen_get_width (screen) / 2)
 			menu_x -= requisition.width;
 		else
@@ -1102,15 +1117,20 @@ panel_applet_list_applets (void)
 }
 
 AppletInfo *
-panel_applet_get_by_type (PanelObjectType object_type)
+panel_applet_get_by_type (PanelObjectType object_type, GdkScreen *screen)
 {
 	GSList *l;
 
 	for (l = registered_applets; l; l = l->next) {
 		AppletInfo *info = l->data;
 
-		if (info->type == object_type)
-			return info;
+		if (info->type == object_type) {
+			if (screen) {
+				if (screen == gtk_widget_get_screen (info->widget))
+					return info;
+			} else
+				return info;
+		}
 	}
 
 	return NULL;
