@@ -1,3 +1,4 @@
+/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*
  * panel-applet-frame.c: panel side container for applets
  *
@@ -27,7 +28,6 @@
 
 #include <libbonoboui.h>
 #include <gconf/gconf.h>
-#include <libgnome/gnome-i18n.h>
 #include <gdk/gdk.h>
 #include <gdk/gdkx.h>
 
@@ -39,6 +39,7 @@
 #include "panel-marshal.h"
 #include "panel-background.h"
 #include "panel-lockdown.h"
+#include "panel-stock-icons.h"
 
 #define PANEL_APPLET_FRAME_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), PANEL_TYPE_APPLET_FRAME, PanelAppletFramePrivate))
 
@@ -46,8 +47,6 @@
 
 #define HANDLE_SIZE 10
 
-#define PANEL_STOCK_DONT_DELETE "panel-dont-delete"
-
 #define PROPERTY_ORIENT     "panel-applet-orient"
 #define PROPERTY_SIZE       "panel-applet-size"
 #define PROPERTY_BACKGROUND "panel-applet-background"
@@ -75,6 +74,14 @@ struct _PanelAppletFramePrivate {
 	guint                           has_handle : 1;
 };
 
+typedef struct {
+	PanelAppletFrame *frame;
+	gboolean          locked;
+	int               position;
+	gboolean          exactpos;
+	char             *id;
+} PanelAppletFrameActivating;
+
 static GObjectClass *parent_class;
 
 /* Keep in sync with panel-applet.h. Uggh.
@@ -89,7 +96,16 @@ typedef enum {
 static void panel_applet_frame_cnx_broken (ORBitConnection  *cnx,
 					   PanelAppletFrame *frame);
 
-static void
+static char *panel_applet_frame_construct_moniker (PanelAppletFrame *frame,
+						   PanelWidget      *panel,
+						   const char       *iid,
+						   const char       *id);
+
+static void panel_applet_frame_activated (Bonobo_Unknown     object,
+					  CORBA_Environment *ev,
+					  gpointer          *data);
+
+void
 panel_applet_frame_sync_menu_state (PanelAppletFrame *frame)
 {
 	PanelWidget *panel_widget;
@@ -108,9 +124,9 @@ panel_applet_frame_sync_menu_state (PanelAppletFrame *frame)
 	locked = panel_widget_get_applet_locked (panel_widget, GTK_WIDGET (frame));
 
 	bonobo_ui_component_set_prop (frame->priv->ui_component,
-				      "/popups/button3/placeholder/lock",
-				      "label",
-				      locked ? _("Un_lock") : _("_Lock"),
+				      "/commands/LockAppletToPanel",
+				      "state",
+				      locked ? "1" : "0",
 				      NULL);
 
 	/* First sensitivity */
@@ -237,10 +253,32 @@ popup_handle_remove (BonoboUIComponent *uic,
 }
 
 static void
-popup_handle_lock (BonoboUIComponent *uic,
-		   PanelAppletFrame  *frame,
-		   const char        *verb)
+listener_popup_handle_lock (BonoboUIComponent            *uic,
+			    const char                   *path,
+			    Bonobo_UIComponent_EventType  type,
+			    const char                   *state,
+			    gpointer                      data)
 {
+	PanelAppletFrame *frame;
+	AppletInfo       *info;
+	gboolean          s;
+
+	g_assert (!strcmp (path, "LockAppletToPanel"));
+
+	if (type != Bonobo_UIComponent_STATE_CHANGED)
+		return;
+
+	if (!state)
+		return;
+
+	frame = (PanelAppletFrame *) data;
+	info = frame->priv->applet_info;
+	s = (strcmp (state, "1") == 0);
+
+	if (panel_widget_get_applet_locked (PANEL_WIDGET (info->widget->parent),
+					    info->widget) == s)
+		return;
+
 	panel_applet_toggle_locked (frame->priv->applet_info);
 
 	panel_applet_frame_sync_menu_state (frame);
@@ -265,13 +303,11 @@ popup_handle_move (BonoboUIComponent *uic,
 
 static BonoboUIVerb popup_verbs [] = {
         BONOBO_UI_UNSAFE_VERB ("RemoveAppletFromPanel", popup_handle_remove),
-	BONOBO_UI_UNSAFE_VERB ("LockAppletToPanel",     popup_handle_lock),
         BONOBO_UI_UNSAFE_VERB ("MoveApplet",            popup_handle_move),
 
         BONOBO_UI_VERB_END
 };
 
-
 static void
 panel_applet_frame_load (const gchar *iid,
 			 PanelWidget *panel,
@@ -280,34 +316,38 @@ panel_applet_frame_load (const gchar *iid,
 			 gboolean     exactpos,
 			 const char  *id)
 {
-	PanelAppletFrame *frame;
-	GtkWidget        *widget;
-	AppletInfo       *info;
+	PanelAppletFrame           *frame;
+	CORBA_Environment           ev;
+	char                       *moniker;
+	PanelAppletFrameActivating *frame_act;
 
 	g_return_if_fail (iid != NULL);
 	g_return_if_fail (panel != NULL);
 	g_return_if_fail (id != NULL);
 
-	widget = panel_applet_frame_new (panel, iid, id);
-	if (!widget)
+	if (panel_lockdown_is_applet_disabled (iid))
 		return;
-	
-	gtk_widget_show_all (widget);
 
-	info = panel_applet_register (widget, widget, NULL,
-				      panel, locked, position,
-				      exactpos, PANEL_OBJECT_BONOBO, id);
+	frame = g_object_new (PANEL_TYPE_APPLET_FRAME, NULL);
+	frame->priv->panel = panel;
+	frame->priv->iid   = g_strdup (iid);
 
-	panel_widget_set_applet_size_constrained (panel, widget, TRUE);
+	frame_act = g_new (PanelAppletFrameActivating, 1);
+	frame_act->frame    = frame;
+	frame_act->locked   = locked;
+	frame_act->position = position;
+	frame_act->exactpos = exactpos;
+	frame_act->id       = g_strdup (id);
 
-	frame = PANEL_APPLET_FRAME (widget);
+	moniker = panel_applet_frame_construct_moniker (frame, panel, iid, id);
 
-	panel_applet_frame_set_info (frame, info);
-	panel_applet_frame_sync_menu_state (frame);
-	panel_applet_frame_init_properties (frame);
+	CORBA_exception_init (&ev);
 
-	panel_lockdown_notify_add (G_CALLBACK (panel_applet_frame_sync_menu_state),
-				   frame);
+	bonobo_get_object_async (moniker, "IDL:Bonobo/Control:1.0", &ev,
+				 (BonoboMonikerAsyncFn) panel_applet_frame_activated,
+				 frame_act);
+
+	g_free (moniker);
 }
 
 void
@@ -316,16 +356,13 @@ panel_applet_frame_load_from_gconf (PanelWidget *panel_widget,
 				    int          position,
 				    const char  *id)
 {
-	const char  *profile;
 	const char  *key;
 	char        *applet_iid;
 
 	g_return_if_fail (panel_widget != NULL);
 	g_return_if_fail (id != NULL);
 
-	profile = panel_profile_get_name ();
-
-	key = panel_gconf_full_key (PANEL_GCONF_APPLETS, profile, id, "bonobo_iid");
+	key = panel_gconf_full_key (PANEL_GCONF_APPLETS, id, "bonobo_iid");
 	applet_iid = gconf_client_get_string (panel_gconf_get_client (), key, NULL);
 
 	panel_applet_frame_load (applet_iid, panel_widget, locked, position, TRUE, id);
@@ -339,18 +376,16 @@ panel_applet_frame_create (PanelToplevel *toplevel,
 			   const char    *iid)
 {
 	GConfClient *client;
-	const char  *profile;
 	const char  *key;
 	char        *id;
 
 	g_return_if_fail (iid != NULL);
 
 	client  = panel_gconf_get_client ();
-	profile = panel_profile_get_name ();
 
 	id = panel_profile_prepare_object (PANEL_OBJECT_BONOBO, toplevel, position, FALSE);
 
-	key = panel_gconf_full_key (PANEL_GCONF_APPLETS, profile, id, "bonobo_iid");
+	key = panel_gconf_full_key (PANEL_GCONF_APPLETS, id, "bonobo_iid");
 	gconf_client_set_string (client, key, iid, NULL);
 
 	panel_profile_add_to_list (PANEL_GCONF_APPLETS, id);
@@ -420,10 +455,31 @@ panel_applet_frame_get_background_string (PanelAppletFrame    *frame,
 					  PanelWidget         *panel,
 					  PanelBackgroundType  type)
 {
-	return panel_background_make_string (
-			&panel->background,
-			GTK_WIDGET (frame)->allocation.x,
-			GTK_WIDGET (frame)->allocation.y);
+	int x;
+	int y;
+
+	x = GTK_WIDGET (frame)->allocation.x;
+	y = GTK_WIDGET (frame)->allocation.y;
+
+	if (frame->priv->has_handle) {
+		switch (frame->priv->orientation) {
+		case PANEL_ORIENTATION_TOP:
+		case PANEL_ORIENTATION_BOTTOM:
+			if (gtk_widget_get_direction (GTK_WIDGET (frame)) !=
+			    GTK_TEXT_DIR_RTL)
+				x += frame->priv->handle_rect.width;
+			break;
+		case PANEL_ORIENTATION_LEFT:
+		case PANEL_ORIENTATION_RIGHT:
+			y += frame->priv->handle_rect.height;
+			break;
+		default:
+			g_assert_not_reached ();
+			break;
+		}
+	}
+
+	return panel_background_make_string (&panel->background, x, y);
 }
 
 void
@@ -447,13 +503,6 @@ panel_applet_frame_change_background (PanelAppletFrame    *frame,
 	}
 }
 
-void
-panel_applet_frame_set_info (PanelAppletFrame *frame,
-			     AppletInfo       *info)
-{
-	frame->priv->applet_info = info;
-}
-
 static void
 panel_applet_frame_finalize (GObject *object)
 {
@@ -645,7 +694,15 @@ panel_applet_frame_size_allocate (GtkWidget     *widget,
 		frame->priv->handle_rect.width  = HANDLE_SIZE;
 		frame->priv->handle_rect.height = allocation->height;
 
-		new_allocation.x      = HANDLE_SIZE;
+		if (gtk_widget_get_direction (GTK_WIDGET (frame)) !=
+		    GTK_TEXT_DIR_RTL) {
+			frame->priv->handle_rect.x = 0;
+			new_allocation.x = HANDLE_SIZE;
+		} else {
+			frame->priv->handle_rect.x = allocation->width - HANDLE_SIZE;
+			new_allocation.x = 0;
+		}
+
 		new_allocation.y      = 0;
 		new_allocation.width  = allocation->width - HANDLE_SIZE;
 		new_allocation.height = allocation->height;
@@ -830,18 +887,19 @@ panel_applet_frame_get_name (char *iid)
 
 	list = bonobo_activation_query (query, NULL, NULL);
 	if (list && list->_length > 0 && list->_buffer) {
-		Bonobo_ServerInfo *info = &list->_buffer [0];
-		const GList       *langs_glist;
-		GSList            *langs_gslist;
+		Bonobo_ServerInfo  *info = &list->_buffer [0];
+		const char * const *langs;
+		GSList             *langs_gslist;
+		int                 i;
+
+		langs = g_get_language_names ();
 
-		/* Evil evil evil evil, we need to convert to
-		 * a GSList from a GList */
-		langs_glist = gnome_i18n_get_language_list ("LC_MESSAGES");
 		langs_gslist = NULL;
-		while (langs_glist) {
-			langs_gslist = g_slist_append (langs_gslist, langs_glist->data);
-			langs_glist = langs_glist->next;
-		}
+		for (i = 0; langs[i]; i++)
+			langs_gslist = g_slist_prepend (langs_gslist,
+							(char *) langs[i]);
+
+		langs_gslist = g_slist_reverse (langs_gslist);
 
 		retval = g_strdup (bonobo_server_info_prop_lookup (
 						info, "name", langs_gslist));
@@ -871,18 +929,20 @@ panel_applet_frame_cnx_broken (ORBitConnection  *cnx,
 	if (frame->priv->iid)
 		applet_name = panel_applet_frame_get_name (frame->priv->iid);
 
-	dialog_txt = g_strconcat ("<span weight=\"bold\" size=\"larger\">",
-				  applet_name ?
-					_("\"%s\" Has Quit Unexpectedly") :
-					_("Panel Object Has Quit Unexpectedly"),
-				  "</span>\n\n",
-				  "If you reload a panel object, it will automatically "
-				  "be added back to the panel.",
-				  NULL);
+	if (applet_name)
+		dialog_txt = g_strdup_printf (_("\"%s\" has quit unexpectedly"), applet_name);
+	else
+		dialog_txt = g_strdup (_("Panel object has quit unexpectedly"));
 
-	dialog = gtk_message_dialog_new_with_markup (NULL, GTK_DIALOG_DESTROY_WITH_PARENT,
-						     GTK_MESSAGE_WARNING, GTK_BUTTONS_NONE,
-						     dialog_txt, applet_name ? applet_name : NULL);
+	dialog = gtk_message_dialog_new (NULL, GTK_DIALOG_DESTROY_WITH_PARENT,
+					 GTK_MESSAGE_WARNING, GTK_BUTTONS_NONE,
+					 dialog_txt, applet_name ? applet_name : NULL);
+
+	gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog),
+						  _("If you reload a panel object, it will automatically "
+						    "be added back to the panel."));
+
+	gtk_container_set_border_width (GTK_CONTAINER (dialog), 6);
 
 	gtk_dialog_add_buttons (GTK_DIALOG (dialog),
 				_("_Don't Reload"), GTK_RESPONSE_NO,
@@ -902,86 +962,45 @@ panel_applet_frame_cnx_broken (ORBitConnection  *cnx,
 	g_free (dialog_txt);
 }
 
-static inline void
-register_stock_item (void)
-{
-        static gboolean registered = FALSE;
-
-        if (!registered) {
-                GtkIconFactory      *factory;
-                GtkIconSet          *cancel_icons;
-
-                static GtkStockItem  dont_delete_item [] = {
-                        { PANEL_STOCK_DONT_DELETE, N_("D_on't Delete"), 0, 0, GETTEXT_PACKAGE },
-                };
-
-                cancel_icons = gtk_icon_factory_lookup_default (GTK_STOCK_CANCEL);
-
-                factory = gtk_icon_factory_new ();
-
-                gtk_icon_factory_add (factory, PANEL_STOCK_DONT_DELETE, cancel_icons);
-
-                gtk_icon_factory_add_default (factory);
-
-                gtk_stock_add_static (dont_delete_item, 1);
-
-                registered = TRUE;
-        }
-}
-
 enum {
 	LOADING_FAILED_RESPONSE_DONT_DELETE,
 	LOADING_FAILED_RESPONSE_DELETE
 };
+ 
+static void
+panel_applet_frame_loading_failed_response (GtkWidget *dialog,
+					    guint      response,
+					    char      *id)
+{
+	gtk_widget_destroy (dialog);
+
+	if (response == LOADING_FAILED_RESPONSE_DELETE &&
+	    !panel_lockdown_get_locked_down () &&
+	    panel_profile_id_lists_are_writable ())
+		panel_profile_remove_from_list (PANEL_GCONF_APPLETS, id);
+
+	g_free (id);
+}
 
 static void
 panel_applet_frame_loading_failed (PanelAppletFrame  *frame,
-				   CORBA_Environment *ev,
-				   const char        *iid,
-				   const char        *id,
-				   GtkWindow         *panel)
+				   const char        *id)
 {
 	GtkWidget *dialog;
-	int        response;
 	char      *problem_txt;
 	gboolean   locked_down;
 
 	locked_down = panel_lockdown_get_locked_down ();
 
 	problem_txt = g_strdup_printf (_("The panel encountered a problem "
-					 "while loading \"%s\"."), iid);
+					 "while loading \"%s\"."),
+				       frame->priv->iid);
 
-	if (ev != NULL) {
-		char *error = bonobo_exception_get_text (ev);
-		char *detail_txt = g_strdup_printf (_("Details: %s"), error);
-
-		dialog = gtk_message_dialog_new (
-					NULL, 0,
-					locked_down ? GTK_MESSAGE_INFO : GTK_MESSAGE_QUESTION,
-					GTK_BUTTONS_NONE,
-					"%s\n%s%s",
-					problem_txt,
-					detail_txt,
-					locked_down ? "" :
-					_("\n\n"
-					  "Do you want to delete the applet "
-					  "from your configuration?"));
-		g_free (error);
-		g_free (detail_txt);
-	} else {
-		dialog = gtk_message_dialog_new (
-					NULL, 0,
-					locked_down ? GTK_MESSAGE_INFO : GTK_MESSAGE_QUESTION,
-					GTK_BUTTONS_NONE,
-					"%s%s",
-					problem_txt,
-					locked_down ? "" :
-					_("\n\n"
-					  "Do you want to delete the applet "
-					  "from your configuration?"));
-	}
-
-	register_stock_item ();
+	dialog = gtk_message_dialog_new (NULL, 0,
+					 locked_down ? GTK_MESSAGE_INFO : GTK_MESSAGE_QUESTION,
+					 GTK_BUTTONS_NONE,
+					 problem_txt);
+	g_free (problem_txt);
 
 	if (locked_down) {
 		gtk_dialog_add_buttons (GTK_DIALOG (dialog),
@@ -989,6 +1008,9 @@ panel_applet_frame_loading_failed (PanelAppletFrame  *frame,
 					NULL);
 		gtk_dialog_set_default_response (GTK_DIALOG (dialog), LOADING_FAILED_RESPONSE_DONT_DELETE);
 	} else {
+		gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog),
+					_("Do you want to delete the applet "
+					  "from your configuration?"));
 		gtk_dialog_add_buttons (GTK_DIALOG (dialog),
 					PANEL_STOCK_DONT_DELETE, LOADING_FAILED_RESPONSE_DONT_DELETE,
 					GTK_STOCK_DELETE, LOADING_FAILED_RESPONSE_DELETE,
@@ -997,16 +1019,16 @@ panel_applet_frame_loading_failed (PanelAppletFrame  *frame,
 	}
 
 	gtk_window_set_screen (GTK_WINDOW (dialog),
-			       gtk_window_get_screen (panel));
+			       gtk_window_get_screen (GTK_WINDOW (frame->priv->panel->toplevel)));
 
-        response = gtk_dialog_run (GTK_DIALOG (dialog));
+	g_signal_connect (dialog, "response",
+			  G_CALLBACK (panel_applet_frame_loading_failed_response),
+			  g_strdup (id));
 
-	gtk_widget_destroy (dialog);
+	gtk_widget_show_all (dialog);
 
-	if (response == LOADING_FAILED_RESPONSE_DELETE &&
-	    !locked_down &&
-	    panel_profile_id_lists_are_writable ())
-		panel_profile_remove_from_list (PANEL_GCONF_APPLETS, id);
+	g_free (frame->priv->iid);
+	g_free (frame);
 }
 
 static void
@@ -1167,9 +1189,9 @@ panel_applet_frame_construct_moniker (PanelAppletFrame *frame,
 	locked_down = panel_lockdown_get_locked_down ();
 
 	retval = g_strdup_printf (
-			"%s!prefs_key=/apps/panel/profiles/%s/applets/%s/prefs;"
+			"%s!prefs_key=/apps/panel/applets/%s/prefs;"
 			"background=%s;orient=%s;size=%s;locked_down=%s",
-			iid, panel_profile_get_name (), id, bg_str,
+			iid, id, bg_str,
 			panel_applet_frame_get_orient_string (frame, panel),
 			panel_applet_frame_get_size_string (frame, panel),
 			locked_down ? "true" : "false");
@@ -1193,80 +1215,81 @@ panel_applet_frame_event_listener (BonoboListener    *listener,
 		panel_applet_frame_set_size_hints_from_any (frame, any);
 }
 
-GtkWidget *
-panel_applet_frame_construct (PanelAppletFrame *frame,
-			      PanelWidget      *panel,
-			      const char       *iid,
-			      const char       *id)
+static void
+panel_applet_frame_activated (Bonobo_Unknown     object,
+			      CORBA_Environment *ev,
+			      gpointer          *data)
 {
+	PanelAppletFrameActivating *frame_act;
+	PanelAppletFrame   *frame;
+	GtkWidget          *widget;
 	BonoboControlFrame *control_frame;
 	Bonobo_Control      control;
-	CORBA_Environment   ev;
-	GtkWidget          *widget;
-	char               *moniker;
+	CORBA_Environment   corba_ev;
+	AppletInfo         *info;
+	char               *error;
 
-	frame->priv->panel = panel;
-
-	if (panel_lockdown_is_applet_disabled (iid))
-		return NULL;
-
-	moniker = panel_applet_frame_construct_moniker (frame, panel, iid, id);
-
-	/* FIXME: this should really use bonobo_get_object_async */
-	CORBA_exception_init (&ev);
-
-	control = bonobo_get_object (
-		moniker, "IDL:Bonobo/Control:1.0", &ev);
-
-	g_free (moniker);
+	frame_act = (PanelAppletFrameActivating *) data;
+	frame = frame_act->frame;
 
 	/* according to the source of bonobo control == NULL && no
 	   exception can happen, so handle it */
-	if (BONOBO_EX (&ev) || control == CORBA_OBJECT_NIL) {
-		panel_applet_frame_loading_failed (
-			frame, &ev, iid, id, GTK_WINDOW (panel->toplevel));
-		CORBA_exception_free (&ev);
-		return NULL;
+	if (BONOBO_EX (ev) || object == CORBA_OBJECT_NIL) {
+		error = bonobo_exception_get_text (ev);
+		g_warning (G_STRLOC ": failed to load applet %s:\n%s",
+			   frame->priv->iid, error);
+		panel_applet_frame_loading_failed (frame, frame_act->id);
+		g_free (frame_act->id);
+		g_free (frame_act);
+		g_free (error);
+		return;
 	}
 
-	frame->priv->iid = g_strdup (iid);
-
-	frame->priv->control = CORBA_Object_duplicate (control, NULL);
-	ORBit_small_listen_for_broken (control,
+	frame->priv->control = CORBA_Object_duplicate (object, NULL);
+	ORBit_small_listen_for_broken (object,
 				       G_CALLBACK (panel_applet_frame_cnx_broken),
 				       frame);
 
-	widget = bonobo_widget_new_control_from_objref (control, CORBA_OBJECT_NIL);
+	widget = bonobo_widget_new_control_from_objref (object,
+							CORBA_OBJECT_NIL);
 
-	bonobo_object_release_unref (control, NULL);
+	bonobo_object_release_unref (object, NULL);
 
 	if (!widget) {
-		CORBA_exception_free (&ev);
-		panel_applet_frame_loading_failed (
-			frame, NULL, iid, id, GTK_WINDOW (panel->toplevel));
-		g_warning (G_STRLOC ": failed to load applet %s", iid);
-		return NULL;
+		g_warning (G_STRLOC ": failed to load applet %s",
+			   frame->priv->iid);
+		panel_applet_frame_loading_failed (frame, frame_act->id);
+		g_free (frame_act->id);
+		g_free (frame_act);
+		return;
 	}
 
 	control_frame = bonobo_widget_get_control_frame (BONOBO_WIDGET (widget));
 	if (control_frame == NULL) {
-		CORBA_exception_free (&ev);
-		panel_applet_frame_loading_failed (
-			frame, NULL, iid, id, GTK_WINDOW (panel->toplevel));
-		g_warning (G_STRLOC ": failed to load applet %s (can't get control frame)", iid);
+		g_warning (G_STRLOC ": failed to load applet %s "
+			   "(can't get control frame)", frame->priv->iid);
+		panel_applet_frame_loading_failed (frame, frame_act->id);
 		gtk_object_sink (GTK_OBJECT (widget));
-		return NULL;
+		g_free (frame_act->id);
+		g_free (frame_act);
+		return;
 	}
 
 	frame->priv->property_bag = 
-		bonobo_control_frame_get_control_property_bag (control_frame, &ev);
-	if (frame->priv->property_bag == NULL || BONOBO_EX (&ev)) {
-		panel_applet_frame_loading_failed (
-			frame, &ev, iid, id, GTK_WINDOW (panel->toplevel));
-		CORBA_exception_free (&ev);
-		g_warning (G_STRLOC ": failed to load applet %s (can't get property bag)", iid);
+		bonobo_control_frame_get_control_property_bag (control_frame,
+							       &corba_ev);
+	if (frame->priv->property_bag == NULL || BONOBO_EX (&corba_ev)) {
+		error = bonobo_exception_get_text (&corba_ev);
+		CORBA_exception_free (&corba_ev);
+		g_warning (G_STRLOC ": failed to load applet %s "
+			   "(can't get property bag):\n%s",
+			   frame->priv->iid, error);
+		panel_applet_frame_loading_failed (frame, frame_act->id);
 		gtk_object_sink (GTK_OBJECT (widget));
-		return NULL;
+		g_free (frame_act->id);
+		g_free (frame_act);
+		g_free (error);
+		return;
 	}
 
 	bonobo_event_source_client_add_listener (frame->priv->property_bag,
@@ -1276,72 +1299,81 @@ panel_applet_frame_construct (PanelAppletFrame *frame,
 						 frame);
 	
 	frame->priv->ui_component =
-		bonobo_control_frame_get_popup_component (control_frame, &ev);
-	if (frame->priv->ui_component == NULL || BONOBO_EX (&ev)) {
-		panel_applet_frame_loading_failed (
-			frame, &ev, iid, id, GTK_WINDOW (panel->toplevel));
-		CORBA_exception_free (&ev);
-		g_warning (G_STRLOC ": failed to load applet %s (can't get popup component)", iid);
+		bonobo_control_frame_get_popup_component (control_frame,
+							  &corba_ev);
+	if (frame->priv->ui_component == NULL || BONOBO_EX (&corba_ev)) {
+		error = bonobo_exception_get_text (&corba_ev);
+		CORBA_exception_free (&corba_ev);
+		g_warning (G_STRLOC ": failed to load applet %s "
+			   "(can't get popup component):\n%s",
+			   frame->priv->iid, error);
+		panel_applet_frame_loading_failed (frame, frame_act->id);
 		gtk_object_sink (GTK_OBJECT (widget));
-		return NULL;
+		g_free (frame_act->id);
+		g_free (frame_act);
+		g_free (error);
+		return;
 	}
 
 	bonobo_ui_util_set_ui (frame->priv->ui_component, DATADIR,
 			       "GNOME_Panel_Popup.xml", "panel", NULL);
 
+	bonobo_ui_component_add_listener (frame->priv->ui_component,
+					  "LockAppletToPanel",
+					  listener_popup_handle_lock,
+					  frame);
+
 	bonobo_ui_component_add_verb_list_with_data (
 		frame->priv->ui_component, popup_verbs, frame);
 
 	control = bonobo_control_frame_get_control (control_frame);
 	if (!control) {
-		CORBA_exception_free (&ev);
-		panel_applet_frame_loading_failed (
-			frame, NULL, iid, id, GTK_WINDOW (panel->toplevel));
-		g_warning (G_STRLOC ": failed to load applet %s (can't get control)", iid);
+		CORBA_exception_free (&corba_ev);
+		g_warning (G_STRLOC ": failed to load applet %s "
+			   "(can't get control)", frame->priv->iid);
+		panel_applet_frame_loading_failed (frame, frame_act->id);
 		gtk_object_sink (GTK_OBJECT (widget));
-		return NULL;
+		g_free (frame_act->id);
+		g_free (frame_act);
+		return;
 	}
 
 	frame->priv->applet_shell = panel_applet_frame_get_applet_shell (control);
 	if (frame->priv->applet_shell == CORBA_OBJECT_NIL) {
-		CORBA_exception_free (&ev);
-		panel_applet_frame_loading_failed (
-			frame, NULL, iid, id, GTK_WINDOW (panel->toplevel));
-		g_warning (G_STRLOC ": failed to load applet %s (can't get applet shell)", iid);
+		CORBA_exception_free (&corba_ev);
+		g_warning (G_STRLOC ": failed to load applet %s "
+			   "(can't get applet shell)", frame->priv->iid);
+		panel_applet_frame_loading_failed (frame, frame_act->id);
 		gtk_object_sink (GTK_OBJECT (widget));
-		return NULL;
+		g_free (frame_act->id);
+		g_free (frame_act);
+		return;
 	}
 
-	CORBA_exception_free (&ev);
+	CORBA_exception_free (&corba_ev);
 
 	gtk_container_add (GTK_CONTAINER (frame), widget);
 
-	return widget;
-}
+	gtk_widget_show_all (GTK_WIDGET (frame));
 
-GtkWidget *
-panel_applet_frame_new (PanelWidget *panel,
-			const char  *iid,
-			const char  *id)
-{
-	PanelAppletFrame *frame;
+	info = panel_applet_register (GTK_WIDGET (frame), GTK_WIDGET (frame),
+				      NULL, frame->priv->panel,
+				      frame_act->locked, frame_act->position,
+				      frame_act->exactpos, PANEL_OBJECT_BONOBO,
+				      frame_act->id);
+	frame->priv->applet_info = info;
 
-	g_return_val_if_fail (iid != NULL && id != NULL, NULL);
+	panel_widget_set_applet_size_constrained (frame->priv->panel,
+						  GTK_WIDGET (frame), TRUE);
 
-	frame = g_object_new (PANEL_TYPE_APPLET_FRAME, NULL);
+	panel_applet_frame_sync_menu_state (frame);
+	panel_applet_frame_init_properties (frame);
 
-	if (!panel_applet_frame_construct (frame, panel, iid, id))
-		return NULL;
+	panel_lockdown_notify_add (G_CALLBACK (panel_applet_frame_sync_menu_state),
+				   frame);
 
-	return GTK_WIDGET (frame);
-}
-
-PanelWidget *
-panel_applet_frame_get_panel (PanelAppletFrame *frame)
-{
-	g_return_val_if_fail (PANEL_IS_APPLET_FRAME (frame), NULL);
-
-	return frame->priv->panel;
+	g_free (frame_act->id);
+	g_free (frame_act);
 }
 
 void
