@@ -519,6 +519,9 @@ load_background_file (PanelBackground *background)
 {
 	GError *error = NULL;
 
+	if (!g_file_test (background->image, G_FILE_TEST_IS_REGULAR))
+		return;
+
 	background->loaded_image = 
 		gdk_pixbuf_new_from_file (background->image, &error);
 	if (!background->loaded_image) {
@@ -882,6 +885,10 @@ panel_background_free (PanelBackground *background)
 		g_object_unref (background->loaded_image);
 	background->loaded_image = NULL;
 
+	if (background->monitor)
+		g_object_unref (background->monitor);
+	background->monitor = NULL;
+
 	if (background->window)
 		g_object_unref (background->window);
 	background->window = NULL;
@@ -904,10 +911,15 @@ panel_background_make_string (PanelBackground *background,
 			      int              x,
 			      int              y)
 {
-	char *retval = NULL;
+	PanelBackgroundType  effective_type;
+	char                *retval;
+	
+	retval = NULL;
 
-	if (background->type == PANEL_BACK_IMAGE ||
-	    (background->type == PANEL_BACK_COLOR && background->has_alpha)) {
+	effective_type = panel_background_effective_type (background);
+
+	if (effective_type == PANEL_BACK_IMAGE ||
+	    (effective_type == PANEL_BACK_COLOR && background->has_alpha)) {
 		GdkNativeWindow pixmap_xid;
 
 		if (!background->pixmap)
@@ -918,7 +930,7 @@ panel_background_make_string (PanelBackground *background,
 
 		retval = g_strdup_printf ("pixmap:%d,%d,%d", pixmap_xid, x, y);
 
-	} else if (background->type == PANEL_BACK_COLOR)
+	} else if (effective_type == PANEL_BACK_COLOR)
 		retval = g_strdup_printf (
 				"color:%.4x%.4x%.4x",
 				background->color.gdk.red,
@@ -931,6 +943,25 @@ panel_background_make_string (PanelBackground *background,
 
 }
 
+PanelBackgroundType
+panel_background_get_type (PanelBackground *background)
+{
+	return background->type;
+}
+
+const PanelColor *
+panel_background_get_color (PanelBackground *background)
+{
+	return &(background->color);
+}
+
+const GdkPixmap *
+panel_background_get_pixmap (PanelBackground *background)
+{
+	return background->pixmap;
+}
+
+
 /* What are we actually rendering - e.g. if we're supposed to
  * be rendering an image, but haven't got a valid image, then
  * we're rendering the default gtk background.
