@@ -404,7 +404,7 @@ static void
 panel_menu_button_recreate_menu (PanelMenuButton *button)
 {
 	if (button->priv->menu)
-		gtk_menu_detach (GTK_MENU (button->priv->menu));
+		gtk_widget_destroy (button->priv->menu);
 	button->priv->menu = NULL;
 }
 
@@ -776,6 +776,7 @@ panel_menu_button_set_menu_path (PanelMenuButton *button,
 		return;
 
 	button->priv->path_root = panel_menu_scheme_to_path_root (scheme);
+	g_free (scheme);
 
 	if (!button->priv->menu_path && (!menu_path || !menu_path [0]))
 		return;
